!---------------------------------------------------------------------!
!                                PHAML                                !
!                                                                     !
! The Parallel Hierarchical Adaptive MultiLevel code for solving      !
! linear elliptic partial differential equations of the form          !
! (PUx)x + (QUy)y + RU = F on 2D polygonal domains with mixed         !
! boundary conditions, and eigenvalue problems where F is lambda*U.   !
!                                                                     !
! PHAML is public domain software.  It was produced as part of work   !
! done by the U.S. Government, and is not subject to copyright in     !
! the United States.                                                  !
!                                                                     !
!     William F. Mitchell                                             !
!     Applied and Computational Mathematics Division                  !
!     National Institute of Standards and Technology                  !
!     william.mitchell@nist.gov                                       !
!     http://math.nist.gov/phaml                                      !
!                                                                     !
!---------------------------------------------------------------------!

module load_balance

!----------------------------------------------------
! This module contains routines for load balancing.
!
! communication tags in this module are of the form 5xx
!----------------------------------------------------

!----------------------------------------------------
! Other modules used are:

use global
use message_passing
use hash_mod
use gridtype_mod
use linsystype_mod
use grid_util
use zoltan_interf
use stopwatch
use refine_elements
use error_estimators
use hp_strategies
use linsys_io

!----------------------------------------------------

implicit none
private
public partition, redistribute, lightweight_set_weights, &
       delete_unowned_elements, set_halo_owners

!----------------------------------------------------
! The following parameters are defined:

! partition assignment for elements that are not in a single partition
integer, parameter :: NO_PARTITION = -1

!----------------------------------------------------

!----------------------------------------------------
! The following types are defined:

!----------------------------------------------------

!----------------------------------------------------
! The following variables are defined:

!----------------------------------------------------

contains

!          ---------
subroutine partition(linsys,grid,procs,lb,refcont,solver_control, &
                     predictive_load_balance,form,balance_what,new_num_part, &
                     export_gid,export_proc,first_call,weights_already_set)
!          ---------

!----------------------------------------------------
! This routine determines the partitioning of the grid.
!----------------------------------------------------

implicit none

!----------------------------------------------------
! Dummy arguments

type (linsys_type), intent(inout) :: linsys
type (grid_type), target, intent(inout) :: grid
type (proc_info), target, intent(in) :: procs
type (Zoltan_Struct), pointer :: lb
type (refine_options), intent(in) :: refcont
type (solver_options), intent(in) :: solver_control
logical, intent(in) :: predictive_load_balance
integer, intent(in) :: form, balance_what, new_num_part
type(hash_key), pointer :: export_gid(:)
integer, pointer :: export_proc(:)
logical, optional :: first_call, weights_already_set

!----------------------------------------------------
! Local variables:

logical :: loc_weights_already_set
!----------------------------------------------------
! Begin executable code

nullify(export_gid,export_proc)
if (my_proc(procs) == MASTER .or. PARALLEL == SEQUENTIAL) return

! start timing the partitioning process

call reset_watch((/ppartition,cppartition/))
call start_watch((/ppartition,tpartition/))

! set the weights

if (present(weights_already_set)) then
   loc_weights_already_set = weights_already_set
else
   loc_weights_already_set = .false.
endif

if (.not. loc_weights_already_set) then
   call set_weights(linsys,grid,predictive_load_balance,balance_what,refcont, &
                    solver_control,procs)
endif

! select method for partitioning.

select case(form)

   case(RTK)
      call reftree_kway(grid,procs,new_num_part,export_gid, &
                        export_proc,first_call)

   case(ZOLTAN_REFTREE, ZOLTAN_RCB, ZOLTAN_OCT, ZOLTAN_RIB, ZOLTAN_METIS, &
        ZOLTAN_HSFC, ZOLTAN_FILE)
      call zoltan_partition(grid,procs,lb,form,export_gid,export_proc, &
                            first_call)

   case default
      call fatal("unknown partition method specified",procs=procs)

end select

! stop timing the partitioning process

call stop_watch((/ppartition,tpartition/))

end subroutine partition

!          -----------
subroutine set_weights(linsys,grid,predictive,balance_what,refcont, &
                       solver_control,procs)
!          -----------

!----------------------------------------------------
! This routine sets the weights for each element
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(linsys_type), intent(inout) :: linsys
type(grid_type), intent(inout) :: grid
logical, intent(in) :: predictive
integer, intent(in) :: balance_what
type (refine_options), intent(in) :: refcont
type (solver_options), intent(in) :: solver_control
type(proc_info), intent(in) :: procs
!----------------------------------------------------
! Local variables:

integer :: assoc_verts(grid%biggest_elem)
integer :: lev, vert, elem, elem_deg(2), side, edge, total, ivert, new_deg
real(my_real) :: global_max_errind, reftol, normsoln, eta, eta_max, R, &
                 max_errind, two_div_log2, loggamma
character(len=1) :: reftype(grid%biggest_elem)
integer :: new_p(2)
logical(small_logical) :: visited_vert(grid%biggest_vert)
!----------------------------------------------------
! Begin executable code

! count the number of associated vertices for each element

assoc_verts = 0
visited_vert = .false.
do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      do ivert=1,VERTICES_PER_ELEMENT
         vert = grid%element(elem)%vertex(ivert)
         if (visited_vert(vert)) cycle
         visited_vert(vert) = .true.
         assoc_verts(grid%vertex(vert)%assoc_elem) = &
            assoc_verts(grid%vertex(vert)%assoc_elem) + 1
      end do
      elem = grid%element(elem)%next
   end do
end do

! compute error indicators if needed and maximum error indicator

if (.not. grid%errind_up2date) then

   call all_error_indicators(grid,refcont%error_estimator,solver_control, &
                             max_errind=max_errind,refine_control=refcont)
   global_max_errind = phaml_global_max(procs,max_errind,501)

else

   global_max_errind = compute_global_max_errind(grid,refcont,procs, &
                                                 still_sequential=.false., &
                                                 no_notrefinable=.true.)
endif

! some constants for predictive load balancing

eta_max = global_max_errind
R = 100000.0_my_real
two_div_log2 = 2.0_my_real/log(2.0_my_real)
loggamma = log(0.125_my_real)

! for predictive load balancing, determine the refinement tolerance for the
! ONE_REF terminations

reftol = 0.0_my_real
if (predictive .and. refcont%refterm == ONE_REF) then
   call get_grid_info(grid,procs,.false.,540, &
                      total_nelem_leaf=total,no_master=.true.)
   reftol = refcont%reftol/sqrt(real(total,my_real))
! Note: for multi-eigenvalue problems and system, just use the first component
!       of the first eigenvalue for normalizing the refinement tolerance
   if (grid%errtype == RELATIVE_ERROR) then
      call norm_solution(linsys,grid,procs,.false.,1,1,energy=normsoln)
      if (normsoln /= 0.0_my_real) reftol = reftol*normsoln
   endif
endif
if (predictive .and. refcont%refterm == ONE_REF_HALF_ERRIND) then
   reftol = global_max_errind/refcont%inc_factor
endif

! for each element ...

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%isleaf .and. grid%element(elem)%iown) then

         elem_deg = grid%element(elem)%degree

         if (predictive) then

! for predictive load balancing, first make a guess as to whether or not this
! element will be refined and whether it will be refined by h or p

! For ONE_REF terminations, check the tolerance to determine if it will be
! refined.  If so, or if it is not a ONE_REF termination, determine h or p.

            new_p = 0
            if ((refcont%refterm /= ONE_REF .and. &
                 refcont%refterm /= ONE_REF_HALF_ERRIND) .or. &
                 maxval(grid%element_errind(elem,:)) > reftol) then
               call mark_reftype_one(elem,grid,refcont,solver_control, &
                                     global_max_errind,reftype,.false.,new_p)
               if (reftype(elem) /=  "n" .and. new_p(1) /= 0) elem_deg = new_p
            else
               reftype(elem) = "n"
            endif

            select case (reftype(elem))

            case ("n")

! if it will not be refined, just assign the weight to be the number of
! whatever we are balancing on associated with this element

               select case (balance_what)

               case (BALANCE_NONE)
                  grid%element(elem)%weight = 0.0_my_real

               case (BALANCE_ELEMENTS)
                  grid%element(elem)%weight = 1.0_my_real

               case (BALANCE_VERTICES)
                  grid%element(elem)%weight = assoc_verts(elem)

               case (BALANCE_EQUATIONS)
                  grid%element(elem)%weight = assoc_verts(elem) + &
                                              element_dof(elem_deg(1))
                  do side=1,EDGES_PER_ELEMENT
                     edge = grid%element(elem)%edge(side)
                     if (grid%edge(edge)%assoc_elem == elem) then
                        grid%element(elem)%weight = grid%element(elem)%weight +&
                           grid%edge(edge)%degree - 1
                     endif
                  end do

               case default
                  ierr = PHAML_INTERNAL_ERROR
                  call fatal("unknown value for balance_what in set_weights")
                  stop

               end select

            case ("h")

! if it will be refined by h, assign the number of entities there should be
! associated with the children

               select case (balance_what)

               case (BALANCE_NONE)
                  grid%element(elem)%weight = 0.0_my_real

               case (BALANCE_ELEMENTS)
                  grid%element(elem)%weight = 2.0_my_real

               case (BALANCE_VERTICES)
                  grid%element(elem)%weight = assoc_verts(elem) + 1

               case (BALANCE_EQUATIONS)
! the equations are: the associated vertices, the new vertex (guess that it
! will be associated, could be wrong), the bubble equations, the equations on
! the new edge, and the edge equations for associated edges (note the base
! will be refined and has twice as many, and guess they will be associated)
                  if (refcont%edge_rule == MINIMUM_RULE) then
                     new_deg = minval(elem_deg)
                  else
                     new_deg = maxval(elem_deg)
                  endif
                  grid%element(elem)%weight = assoc_verts(elem)+1 + &
                                              element_dof(elem_deg(1)) + &
                                              element_dof(elem_deg(2)) + &
                                              new_deg-1
                  do side=1,EDGES_PER_ELEMENT
                     edge = grid%element(elem)%edge(side)
                     if (grid%edge(edge)%assoc_elem == elem) then
                        if (side == 1) then
                           if (refcont%edge_rule == MINIMUM_RULE) then
                              new_deg = min(grid%edge(edge)%degree,elem_deg(2))
                           else
                              new_deg = max(grid%edge(edge)%degree,elem_deg(2))
                           endif
                           grid%element(elem)%weight = &
                                                   grid%element(elem)%weight + &
                                                   new_deg - 1
                        elseif (side == 2) then
                           if (refcont%edge_rule == MINIMUM_RULE) then
                              new_deg = min(grid%edge(edge)%degree,elem_deg(1))
                           else
                              new_deg = max(grid%edge(edge)%degree,elem_deg(1))
                           endif
                           grid%element(elem)%weight = &
                                                   grid%element(elem)%weight + &
                                                   new_deg - 1
                        else
                           if (refcont%edge_rule == MINIMUM_RULE) then
                              new_deg = min(grid%edge(edge)%degree,elem_deg(1))
                           else
                              new_deg = max(grid%edge(edge)%degree,elem_deg(1))
                           endif
                           grid%element(elem)%weight = &
                                                   grid%element(elem)%weight + &
                                                   new_deg - 1
                           if (refcont%edge_rule == MINIMUM_RULE) then
                              new_deg = min(grid%edge(edge)%degree,elem_deg(2))
                           else
                              new_deg = max(grid%edge(edge)%degree,elem_deg(2))
                           endif
                           grid%element(elem)%weight = &
                                                   grid%element(elem)%weight + &
                                                   new_deg - 1
                        endif
                     endif
                  end do

               case default
                  ierr = PHAML_INTERNAL_ERROR
                  call fatal("unknown value for balance_what in set_weights")
                  stop

               end select

            case ("p")

! if it will be refined by p, assign the number of entities there should be
! associated after p refinement

               if (new_p(1) == 0) elem_deg = elem_deg + 1

               select case (balance_what)

               case (BALANCE_NONE)
                  grid%element(elem)%weight = 0.0_my_real

               case (BALANCE_ELEMENTS)
                  grid%element(elem)%weight = 1.0_my_real

               case (BALANCE_VERTICES)
                  grid%element(elem)%weight = assoc_verts(elem)

               case (BALANCE_EQUATIONS)
! for associated edges, guess the edge degree is the same as this element
                  grid%element(elem)%weight = assoc_verts(elem) + &
                                              element_dof(elem_deg(1))
                  do side=1,EDGES_PER_ELEMENT
                     edge = grid%element(elem)%edge(side)
                     if (grid%edge(edge)%assoc_elem == elem) then
                        grid%element(elem)%weight = grid%element(elem)%weight +&
                                                    elem_deg(1)-1
                     endif
                  end do

               case default
                  ierr = PHAML_INTERNAL_ERROR
                  call fatal("unknown value for balance_what in set_weights")
                  stop

               end select

            case default
               ierr = PHAML_INTERNAL_ERROR
               call fatal("unknown value for reftype in set_weights")
               stop

            end select

! For terminations that are not ONE_REF, multiply by a factor that will be
! larger in elements we expect to be refined more times.  For simplicity,
! we just multiply by an estimate of how many times we think the element
! would have to be refined to reach a certain error, rather than properly
! estimating the number of entities there would be after that many refinements.
! Given: the error estimate for the element, eta
!        the maximum error estimate over all elements, eta_max
!        some artificial factor by which to reduce the maximum error, R
!        a factor by which we believe p refinement reduces the error, gamma
!        the current element degree, p
! the number of refinements needed to reduce the error to eta_max/R is
! ceiling(A log(R*eta/eta_max)) where A is 2/(p*log(2)) for h refinement
! and -log(gamma) for p refinement.
! If eta is already less than eta_max/R, don't multiply by the factor.
! And don't worry about the ceiling.

            if (refcont%refterm /= ONE_REF .and. &
                 refcont%refterm /= ONE_REF_HALF_ERRIND) then
               eta = maxval(grid%element_errind(elem,:))
               if (eta > eta_max/R) then
                  if (reftype(elem) == "h") then
                     grid%element(elem)%weight = grid%element(elem)%weight * &
                      two_div_log2*log(eta*R/eta_max)/grid%element(elem)%degree
                  elseif (reftype(elem) == "p") then
                     grid%element(elem)%weight = grid%element(elem)%weight * &
                      (-loggamma)*log(eta*R/eta_max)
                  endif
               endif
            endif

         else ! not predictive

! if it is not predictive load balancing, just assign the weight to be
! the number of whatever we are balancing on associated with this element

            select case (balance_what)

            case (BALANCE_NONE)
               grid%element(elem)%weight = 0.0_my_real

            case (BALANCE_ELEMENTS)
               grid%element(elem)%weight = 1.0_my_real

            case (BALANCE_VERTICES)
               grid%element(elem)%weight = assoc_verts(elem)

            case (BALANCE_EQUATIONS)
               grid%element(elem)%weight = assoc_verts(elem) + &
                                           element_dof(elem_deg(1))
               do side=1,EDGES_PER_ELEMENT
                  edge = grid%element(elem)%edge(side)
                  if (grid%edge(edge)%assoc_elem == elem) then
                     grid%element(elem)%weight = grid%element(elem)%weight + &
                        grid%edge(edge)%degree - 1
                  endif
               end do

            case default
               ierr = PHAML_INTERNAL_ERROR
               call fatal("unknown value for balance_what in set_weights")
               stop

            end select

         endif ! predictive

      else ! not a leaf or I don't own it

         grid%element(elem)%weight = 0.0_my_real

      endif

      elem = grid%element(elem)%next
   end do
end do

end subroutine set_weights

!          -----------------------
subroutine lightweight_set_weights(grid,balance_what,desired_level, &
                                   desired_degree)
!          -----------------------

!----------------------------------------------------
! This routine sets the weights for each element.  It is a light weight version
! of set_weights in which the h and p refinement of the elements has already
! been determined and is only used for predictive load balancing.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: balance_what
integer, intent(in) :: desired_level(:), desired_degree(:,:)
!----------------------------------------------------
! Local variables:

integer :: assoc_verts(grid%biggest_elem)
integer :: lev, vert, elem, side, edge, delta, p, ivert
logical(small_logical) :: visited_vert(grid%biggest_vert)
!----------------------------------------------------
! Begin executable code

! count the number of associated vertices for each element

assoc_verts = 0
visited_vert = .false.
do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      do ivert=1,VERTICES_PER_ELEMENT
         vert = grid%element(elem)%vertex(ivert)
         if (visited_vert(vert)) cycle
         visited_vert(vert) = .true.
         assoc_verts(grid%vertex(vert)%assoc_elem) = &
            assoc_verts(grid%vertex(vert)%assoc_elem) + 1
      end do
      elem = grid%element(elem)%next
   end do
end do

! for each element ...

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%isleaf .and. grid%element(elem)%iown) then

! Estimate the number of associated vertices, associated edges, elements
! and equations for this element after it is refined to
! (desired_level,desired_degree(1)).
! These formulas are derived for refinements or no change; assume they're
! OK for coarsening, too, but don't set a negative weight.

! Let d (for delta) be the change in refinement level and p be the new degree.

! elements: 2^d

! vertices: it's a 50-50 shot as to whether a new vertex on the element
! boundary will be associated with this element or the neighboring element,
! so count those vertices as 1/2 and interior vertices as 1.
! 2^(d-1) - 1/2 + number currently associated

! edges: the child edges of an edge of the element will be associated with the
! element if and only if the edge is associated, so use that association
! status to determine which ones to include.  The number of edges in the
! refined element depend whether d is even or odd and whether the edge is
! interior or a decendent of the base or a non-based side.
! d odd,  non-base: 2^((d-1)/2)
! d odd,  base:     2^((d+2)/2)
! d odd,  interior: 3*2^(d-1) - 2^((d+1)/2) (zero if negative)
! d even, non-base: 2^d
! d even, base:     2^d
! d even, interior: 3*2^(d-1) - 3*2^(d/2-1) (zero if negative)

! equations: for simplicity, assume the degree of the element edges is the
! same as the degree of this element.  Then the number of equations is the
! number of associated vertices plus the (p-1)*number of descendent edges for
! each associated edge and interior edges plus (p-1)(p-2)/2 times the number
! of descendent elements.

         delta = desired_level(elem) - grid%element(elem)%level
         p = desired_degree(1,elem)

         select case (balance_what)

         case (BALANCE_NONE)
            grid%element(elem)%weight = 0.0_my_real

         case (BALANCE_ELEMENTS)
            grid%element(elem)%weight = 2**delta

         case (BALANCE_VERTICES)
            grid%element(elem)%weight = 2**(delta-1) - 0.5_my_real + &
                                        assoc_verts(elem)

         case (BALANCE_EQUATIONS)
            grid%element(elem)%weight = assoc_verts(elem) + &
                                        2**delta*element_dof(p)
            if (p > 1) then
               if (2*(delta/2) == delta) then
                  do side=1,EDGES_PER_ELEMENT
                     edge = grid%element(elem)%edge(side)
                     if (grid%edge(edge)%assoc_elem == elem) then
                        grid%element(elem)%weight = grid%element(elem)%weight +&
                                                    (p-1)*2**delta
                     endif
                  end do
                  grid%element(elem)%weight = grid%element(elem)%weight + &
                     max(0.0_my_real, &
                     (3.0_my_real*2**(delta-1)-3*2**(delta/2-1))*element_dof(p))
               else
                  do side=1,EDGES_PER_ELEMENT
                     edge = grid%element(elem)%edge(side)
                     if (grid%edge(edge)%assoc_elem == elem) then
                        grid%element(elem)%weight = grid%element(elem)%weight +&
                                                    (p-1)*2**((delta-1)/2)
                        if (side==3) then
                        grid%element(elem)%weight = grid%element(elem)%weight +&
                                                    (p-1)*2**((delta-1)/2)
                        endif
                     endif
                  end do
                  grid%element(elem)%weight = grid%element(elem)%weight + &
                     max(0.0_my_real, &
                     (3.0_my_real*2**(delta-1)-2**((delta+1)/2))*element_dof(p))
               endif
            endif

         case default
            ierr = PHAML_INTERNAL_ERROR
            call fatal("unknown value for balance_what in lightweight_set_weights")
            stop

         end select

      else ! not a leaf or I don't own it

         grid%element(elem)%weight = 0.0_my_real

      endif

      elem = grid%element(elem)%next
   end do
end do

end subroutine lightweight_set_weights

!          ------------
subroutine reftree_kway(grid,procs,new_num_part,export_gid, &
                        export_part,first_call)
!          ------------

!----------------------------------------------------
! This routine partitions the grid using the k-way refinement tree method
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type (grid_type), intent(in) :: grid
type (proc_info), intent(in) :: procs
integer, intent(in) :: new_num_part
type(hash_key), pointer :: export_gid(:)
integer, pointer :: export_part(:)
logical, optional :: first_call
!----------------------------------------------------
! Local variables:

real(my_real) :: root_weight
real(my_real), allocatable :: weight(:)
integer :: elem, i, my_lid, part, num_exp, allocstat, count, nproc, &
           issub, rssub, oissub, orssub, irsub, rrsub
real(my_real) :: part_size, current_size, cutoff
! newcomm
integer :: nisend
integer, allocatable :: isend(:),nisendv(:),nirecv(:),nrsendv(:),nrrecv(:)
integer, pointer :: irecv(:)
real(my_real), allocatable :: rsend(:)
real(my_real), pointer :: rrecv(:)
type(hash_key), allocatable :: leaf_list(:)
! 1 if all descendents are mine, 0 if no descendents are mine, -1 if some
! descendents are mine
integer, allocatable :: mine(:)
integer, allocatable :: new_partition(:)
logical :: skip_comm
!----------------------------------------------------
! Begin executable code

nproc = num_proc(procs)

allocate(nisendv(nproc),nirecv(nproc),nrsendv(nproc),nrrecv(nproc), &
         stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in reftree_kway",procs=procs)
   return
endif

allocate(mine(grid%biggest_elem),weight(grid%biggest_elem),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in reftree_kway",procs=procs)
   return
endif

! sum the weights of the nodes that are assigned to this partition,
! and count the leaves that are not

count = 0
elem = grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   call sum_my_weights(grid,elem,weight,mine,count)
   elem = grid%element(elem)%next
end do

! on first call, all nodes are assigned to all partitions, so skip the
! communication part

if (present(first_call)) then
   if (first_call) then
      if (count == 0) then
         skip_comm = .true.
      else
         call warning("On first call to reftree_kway, found leaves not assigned to this processor.", &
                      "Partitions may be wrong.")
         skip_comm = .true.
      endif
   else
      skip_comm = .false.
   endif
else
   skip_comm = .false.
endif

if (.not. skip_comm) then

! make a list of the leaves that are not assigned to each partition

   allocate(leaf_list(count),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in reftree_kway",procs=procs)
      return
   endif
   count = 0
   elem = grid%head_level_elem(1)
   do while (elem /= END_OF_LIST)
      call list_other_leaves(grid,elem,leaf_list,count)
      elem = grid%element(elem)%next
   end do

! get unknown leaf weights from other partitions

! copy leaf list to isend

   nisend = size(leaf_list)*KEY_SIZE
   allocate(isend(nisend),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in reftree_kway",procs=procs)
      return
   endif
   do i=1,size(leaf_list)
      call hash_pack_key(leaf_list(i),isend,1+(i-1)*KEY_SIZE)
   end do

! send my list of leaves for which I need partial sums, and get request list
! of other partitions

   call start_watch((/cppartition,ctpartition/))
   call phaml_alltoall(procs,isend,nisend,irecv,nirecv,510)
   call stop_watch((/cppartition,ctpartition/))

   deallocate(leaf_list,stat=allocstat)
   deallocate(isend,stat=allocstat)

! count the number of requests for which I have an answer, and allocate
! space for the next message

   count = 0
   irsub = 1
   do part=1,nproc
      do i=1,nirecv(part)/KEY_SIZE
         my_lid = hash_decode_key(hash_unpack_key(irecv,irsub),grid%elem_hash)
         if (my_lid /= HASH_NOT_FOUND) then
            if (weight(my_lid) /= 0.0_my_real) then
               count = count + 1
            endif
         endif
         irsub = irsub + KEY_SIZE
      end do
   end do

   allocate(isend(count*KEY_SIZE),rsend(count),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in reftree_kway",procs=procs)
      return
   endif

! build my reply to the requests

   irsub = 1
   issub = 1
   rssub = 1
   oissub = issub
   orssub = rssub
   do part=1,nproc
      do i=1,nirecv(part)/KEY_SIZE
         my_lid = hash_decode_key(hash_unpack_key(irecv,irsub),grid%elem_hash)
         if (my_lid /= HASH_NOT_FOUND) then
            if (weight(my_lid) /= 0.0_my_real) then
               isend(issub:issub+KEY_SIZE-1) = irecv(irsub:irsub+KEY_SIZE-1)
               issub = issub + KEY_SIZE
               rsend(rssub) = weight(my_lid)
               rssub = rssub + 1
            endif
         endif
         irsub = irsub + KEY_SIZE
      end do
      nisendv(part) = issub - oissub
      nrsendv(part) = rssub - orssub
      oissub = issub
      orssub = rssub
   end do

   if (associated(irecv)) deallocate(irecv,stat=allocstat)

! send the reply

   call start_watch((/cppartition,ctpartition/))
   call phaml_alltoall(procs,isend,nisendv,irecv,nirecv,520)
   call phaml_alltoall(procs,rsend,nrsendv,rrecv,nrrecv,521)
   call stop_watch((/cppartition,ctpartition/))
   deallocate(isend,rsend,stat=allocstat)

! add the replies into my weights

   irsub = 1
   rrsub = 1
   do part=1,nproc
      do i=1,nrrecv(part)
         my_lid = hash_decode_key(hash_unpack_key(irecv,irsub),grid%elem_hash)
         if (my_lid /= HASH_NOT_FOUND) then
            weight(my_lid) = weight(my_lid) + rrecv(rrsub)
         endif
         irsub = irsub + KEY_SIZE
         rrsub = rrsub + 1
      end do
   end do

! finish summing the weights

   elem = grid%head_level_elem(1)
   do while (elem /= END_OF_LIST)
      call sum_all_weights(grid,elem,weight,mine)
      elem = grid%element(elem)%next
   end do

   if (associated(irecv)) deallocate(irecv,stat=allocstat)
   if (associated(rrecv)) deallocate(rrecv,stat=allocstat)
   deallocate(nisendv,nirecv,nrsendv,nrrecv,stat=allocstat)

endif

! compute the root weight

root_weight = 0.0_my_real

elem = grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   root_weight = root_weight + weight(elem)
   elem = grid%element(elem)%next
end do

! determine partition sizes

part_size = root_weight/new_num_part

allocate(new_partition(grid%biggest_elem),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("allocation failed in reftree_kway",procs=procs)
   return
endif

! traverse the tree to define partition and count the number of exports

num_exp = 0
part = 1
current_size = 0.0_my_real
cutoff = part_size

elem = grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   call define_partition(grid,elem,new_partition,weight, &
                         num_exp,part,current_size,cutoff,part_size, &
                         new_num_part,mine)
   elem = grid%element(elem)%next
end do

! if no exports, we're done

if (num_exp == 0) then
   nullify(export_gid, export_part)
else

! allocate space for the export lists

   allocate(export_gid(num_exp), export_part(num_exp), stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in reftree_kway",procs=procs)
      return
   endif

! traverse the tree to make the export lists

   elem = grid%head_level_elem(1)
   count = 0
   do while (elem /= END_OF_LIST)
     call make_export_lists(grid,elem,mine,new_partition,export_gid, &
                            export_part,count)
     elem = grid%element(elem)%next
   end do

endif

deallocate(new_partition,weight,mine,stat=allocstat)

end subroutine reftree_kway

!                    --------------
recursive subroutine sum_my_weights(grid,subroot,weight,mine,count)
!                    --------------

!----------------------------------------------------
! This routine sums the weights of the elements owned by this partition, and
! counts the number of unowned leaves
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: subroot
real(my_real), intent(inout) :: weight(:)
integer, intent(inout) :: mine(:)
integer, intent(inout) :: count
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
logical :: none_assigned, all_assigned
!----------------------------------------------------
! Begin executable code

allc = ALL_CHILDREN
children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)

if (children(1) == NO_CHILD) then

! if there are no children, then the sum is the weight of this node if it
! is assigned to this processor and 0.0 otherwise

   if (grid%element(subroot)%iown) then

      weight(subroot) = grid%element(subroot)%weight
      mine(subroot) = 1
   else
      weight(subroot) = 0.0_my_real
      mine(subroot) = 0
      count = count + 1
   endif

else

! if there are children, sum the weights of the children

   weight(subroot) = 0.0_my_real
   none_assigned = .true.
   all_assigned = .true.
   do i=1,MAX_CHILD
      if (children(i) /= NO_CHILD) then
         call sum_my_weights(grid,children(i),weight,mine,count)
         weight(subroot) = weight(subroot) + weight(children(i))
         if (mine(children(i))==1 .or. mine(children(i))==-1) then
            none_assigned = .false.
         endif
         if (mine(children(i))==0 .or. mine(children(i))==-1) then
            all_assigned = .false.
         endif
      endif
   end do
   if (none_assigned) then
      mine(subroot) = 0
   elseif (all_assigned) then
      mine(subroot) = 1
   else
      mine(subroot) = -1
   endif

endif

end subroutine sum_my_weights

!                    -----------------
recursive subroutine list_other_leaves(grid,subroot,leaf_list,count)
!                    -----------------

!----------------------------------------------------
! This routine creates a list of all the leaves that are not owned by this
! partition
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: subroot
type(hash_key), intent(inout) :: leaf_list(:)
integer, intent(inout) :: count
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

allc = ALL_CHILDREN
children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)

if (children(1) == NO_CHILD) then

   if (.not. grid%element(subroot)%iown) then
      count = count + 1
      leaf_list(count) = grid%element(subroot)%gid
   endif

else

   do i=1,MAX_CHILD
      if (children(i) /= NO_CHILD) then
         call list_other_leaves(grid,children(i),leaf_list,count)
      endif
   end do

endif

end subroutine list_other_leaves

!                    ---------------
recursive subroutine sum_all_weights(grid,subroot,weight,mine)
!                    ---------------

!----------------------------------------------------
! This routine sums the weights of all elements.  Elements for which mine is
! true already have the final sum; unowned leaves have received the sum from
! other partitions.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: subroot
real(my_real), intent(inout) :: weight(:)
integer, intent(inout) :: mine(:)
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

allc = ALL_CHILDREN
children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)

! if no children or I own all descendents, then done

if (children(1) /= NO_CHILD .and. mine(subroot) /= 1) then

! if there are children, sum the weights of the children

   weight(subroot) = 0.0_my_real
   do i=1,MAX_CHILD
      if (children(i) /= NO_CHILD) then
         call sum_all_weights(grid,children(i),weight,mine)
         weight(subroot) = weight(subroot) + weight(children(i))
      endif
   end do

endif

end subroutine sum_all_weights

!                    ----------------
recursive subroutine define_partition(grid,subroot,new_partition, &
                                      weight,num_exp,part,current_size,cutoff, &
                                      partition_size,new_num_part,mine)
!                    ----------------

!----------------------------------------------------
! This routine defines the partition and counts the number of exports
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type (grid_type), intent(in) :: grid
integer, intent(in) :: subroot
integer, intent(inout) :: new_partition(:)
real(my_real), intent(in) :: weight(:)
integer, intent(inout) :: num_exp, part
real(my_real), intent(inout) :: current_size, cutoff
real(my_real), intent(in) :: partition_size
integer, intent(in) :: new_num_part,mine(:)

!----------------------------------------------------
! Local variables:

real(my_real) :: newsize
integer :: children(MAX_CHILD), i, allc(MAX_CHILD)

!----------------------------------------------------
! Begin executable code

newsize = current_size + weight(subroot)
allc = ALL_CHILDREN
children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)

if (newsize <= cutoff .or. part == new_num_part .or. &
    abs(newsize-cutoff)/cutoff < 100*epsilon(0.0_my_real)) then

! this subtree fits in the current partition

   new_partition(subroot) = part
   current_size = newsize

! If this is this processor's partition, there are no exports below this
! node, so we don't have to traverse the subtree.
! If there are no leaves of this subtree assigned to this processor, there
! are no exports below this node.
! Otherwise, traverse the subtree setting partition and counting exports

   if (part /= grid%partition .and. mine(subroot) /= 0) then
      if (children(1) == NO_CHILD) then
         num_exp = num_exp + 1
      else
         do i=1,MAX_CHILD
           call mark_and_count(grid,children(i),part,mine,new_partition,num_exp)
         end do
      endif
   endif

else

! this subtree is too big for the current partition

   if (children(1) /= NO_CHILD) then

! if there are children, traverse them

      new_partition(subroot) = NO_PARTITION
      do i=1,MAX_CHILD
         call define_partition(grid, &
                               children(grid%element(subroot)%order(i)), &
                               new_partition,weight,num_exp,part, &
                               current_size,cutoff,partition_size, &
                               new_num_part,mine)
      end do

! if there are no children, move on to next partition

   else

      do while (newsize > cutoff)
         part = part + 1
         cutoff = part*partition_size
      end do
      new_partition(subroot) = part
      current_size = newsize
      if (part /= grid%partition .and. mine(subroot) /= 0) then
         num_exp = num_exp + 1
      endif
   endif
endif

end subroutine define_partition

!                    --------------
recursive subroutine mark_and_count(grid,subroot,part,mine,new_partition, &
                                    num_exp)
!                    --------------

!----------------------------------------------------
! This routine defines the new partition for elements in a subtree assigned
! to a different partition, and counts the exports
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: subroot,part,mine(:)
integer, intent(inout) :: new_partition(:),num_exp
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

new_partition(subroot) = part
allc = ALL_CHILDREN
children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)

if (children(1) == NO_CHILD) then
   if (mine(subroot) /= 0) num_exp = num_exp + 1
else
   do i=1,MAX_CHILD
      if (children(i) /= NO_CHILD) then
         call mark_and_count(grid,children(i),part,mine,new_partition,num_exp)
      endif
   end do

endif

end subroutine mark_and_count

!                    -----------------
recursive subroutine make_export_lists(grid,subroot,mine,new_partition, &
                                       export_gid,export_proc,count)
!                    -----------------

!----------------------------------------------------
! This routine creates the exports lists
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(in) :: grid
integer, intent(in) :: subroot,mine(:),new_partition(:)
type(hash_key), intent(inout) :: export_gid(:)
integer, intent(inout) :: export_proc(:),count

!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

! if this subtree has no leaves assigned to this processor, or if the
! new partition of this subtree is this processor, then there can be no
! exports below it

if (mine(subroot) /= 0 .and. new_partition(subroot) /= grid%partition) then
   allc = ALL_CHILDREN
   children = get_child_lid(grid%element(subroot)%gid,allc,grid%elem_hash)
   if (children(1) == NO_CHILD) then

! if this is a leaf, put it on the export lists

      count = count + 1
      export_gid(count) = grid%element(subroot)%gid
      export_proc(count) = new_partition(subroot)

   else

! if it is not a leaf, traverse the children

      do i=1,MAX_CHILD
         call make_export_lists(grid,children(i),mine,new_partition, &
                                export_gid,export_proc,count)
      end do
   endif
endif

end subroutine make_export_lists

!          ------------
subroutine redistribute(grid,procs,refine_control,solver_control,export_gid, &
                        export_part,desired_level,desired_degree, &
                        was_hcoarsened,first_call)
!          ------------

!----------------------------------------------------
! This routine redistributes the data structures over the partitions
! after 'partition' determines a new distribution.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type (grid_type), intent(inout) :: grid
type (proc_info), intent(inout) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
type(hash_key), pointer :: export_gid(:)
integer, pointer :: export_part(:)
integer, pointer, optional :: desired_level(:), desired_degree(:,:)
logical(small_logical), pointer, optional :: was_hcoarsened(:)
logical, intent(in), optional :: first_call
!----------------------------------------------------

!----------------------------------------------------
! Local variables:

integer, allocatable :: isend(:), nisend(:), nrsend(:), nirecv(:), nrrecv(:), &
                        iind(:), rind(:)
integer, allocatable :: istart(:),iend(:),rstart(:),rend(:)
integer, pointer :: irecv(:)
real(my_real), allocatable :: rsend(:)
real(my_real), pointer :: rrecv(:)
type(hash_key) :: elemgid, gid
integer i, j, k, l, m, part, elemlid, allocstat, nproc, iindex, &
        rindex, d, to, elem, edge, vert, nerrind, limit, myproc, nexport_halo, &
        ni, nhalo, astat, nneigh
integer :: halo(grid%nelem_leaf)
integer, allocatable :: export_lid(:), export_halo(:), neighbor_procs(:), &
                        inv_neighproc(:)
integer :: newowner(grid%biggest_elem)
integer, pointer :: neighproc(:), allneigh(:)
logical :: skip_comm, need_alltoall
logical, allocatable :: is_neighbor(:)

!----------------------------------------------------
! Begin executable code

! no redistribution if not running in parallel

if (PARALLEL == SEQUENTIAL) return

grid_changed = .true.

! master does not participate

myproc = my_proc(procs)
if (myproc == MASTER) return

! start timing the redistribution process

call reset_watch((/pdistribute,cpdistribute/))
call start_watch((/pdistribute,tdistribute/))

nproc = num_proc(procs)
nerrind = max(1,grid%num_eval)

! is this claimed to the be the first call?

if (present(first_call)) then
   if (first_call) then
      skip_comm = .true.
   else
      skip_comm = .false.
   endif
else
   skip_comm = .false.
endif

if (.not.associated(export_gid)) then
   allocate(export_gid(0),export_part(0),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in redistribute",procs=procs)
      return
   endif
endif

allocate(export_lid(size(export_gid)),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("redistribute: allocation failed",procs=procs)
   stop
endif

! get the lids of the export elements and note the new owner of exported
! elements

newowner = UNKNOWN_OWNER
do i=1,size(export_gid)
   export_lid(i) = hash_decode_key(export_gid(i),grid%elem_hash)
   newowner(export_lid(i)) = export_part(i)
end do

ni = 0
do i=1,size(export_gid)
   call get_all_neighbors(grid,export_lid(i),allneigh)
   ni = ni + size(allneigh)
   deallocate(allneigh)
end do

allocate(export_halo(ni),stat=allocstat)
if (allocstat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("redistribute: allocation failed",procs=procs)
   stop
endif

! remove myself as owner of the elements to be transfered

do i=1,size(export_gid)
   elemlid = export_lid(i)
   grid%element(elemlid)%iown = .false.
   grid%element(elemlid)%owner = UNKNOWN_OWNER
   call disown_parent(grid,export_gid(i))
   if (grid%element(elemlid)%degree >= 3) then
      grid%dof_own = grid%dof_own - grid%system_size * &
           element_dof(grid%element(elemlid)%degree)
   endif
   do j=1,VERTICES_PER_ELEMENT
      if (grid%vertex(grid%element(elemlid)%vertex(j))%assoc_elem == elemlid) then
         grid%nvert_own = grid%nvert_own - 1
         if (is_periodic_vert(grid%element(elemlid)%vertex(j),grid)) then
            grid%nvert_own = grid%nvert_own - &
               num_matching_periodic_vert(grid,grid%element(elemlid)%vertex(j))
         endif
         grid%dof_own = grid%dof_own - grid%system_size
      endif
   end do
   do j=1,EDGES_PER_ELEMENT
      if (grid%edge(grid%element(elemlid)%edge(j))%assoc_elem == elemlid) then
         grid%nedge_leaf_own = grid%nedge_leaf_own - 1
         if (is_periodic_edge(grid%element(elemlid)%edge(j),grid)) then
            grid%nedge_leaf_own = grid%nedge_leaf_own - 1
         endif
         if (grid%edge(grid%element(elemlid)%edge(j))%degree >= 2) then
            grid%dof_own = grid%dof_own - grid%system_size * &
                       (grid%edge(grid%element(elemlid)%edge(j))%degree - 1)
         endif
      endif
   end do
end do
grid%nelem_leaf_own = grid%nelem_leaf_own - size(export_gid)

! on first call, all processors have all elements, so do not need to
! send the exported elements

if (.not. skip_comm) then

! Get the neighbor processors

   call get_neighbor_procs(procs,neighproc)
   if (associated(neighproc)) then
      nneigh = size(neighproc)
   else
      nneigh = 0
   endif
   allocate(inv_neighproc(nproc),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("redistribute: allocation failed",procs=procs)
      stop
   endif
   inv_neighproc = -1
   do i=1,nneigh
      inv_neighproc(neighproc(i)) = i
   end do

! Determine if any element is exported outside the current neighborhood.
! If so, use full alltoall communication.

   j = 0
   do i=1,size(export_part)
      if (inv_neighproc(export_part(i)) == -1) then
         j = 1
         exit
      endif
   end do
   need_alltoall = phaml_global_max(procs,j,532)==1

! update the owner of halo elements

   call update_halo_owners(grid,procs,export_lid,export_part,inv_neighproc, &
                           nneigh,need_alltoall,export_halo,nexport_halo)

! pass through the export list to count how much will be sent to each
! processor.  see comments in unpack_elements

   allocate(nisend(nproc),nrsend(nproc),nirecv(nproc),nrrecv(nproc), &
            istart(nproc),iend(nproc),rstart(nproc),rend(nproc), &
            iind(nproc),rind(nproc),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in redistribute",procs=procs)
      return
   endif

   nisend = 0
   nrsend = 0

   do i=1,size(export_gid)
      elemgid = export_gid(i)
      elemlid = export_lid(i)
      to = export_part(i)
      if (.not. need_alltoall) then
         to = inv_neighproc(to)
      endif
      nisend(to) = nisend(to) + KEY_SIZE
      call get_all_neighbors(grid,elemlid,allneigh)
      nisend(to) = nisend(to) + size(allneigh)*(KEY_SIZE+1) + 1
      deallocate(allneigh)
      nisend(to) = nisend(to) + 4
      nrsend(to) = nrsend(to) + nerrind
      nrsend(to) = nrsend(to) + VERTICES_PER_ELEMENT*grid%nsoln
      do j=1,EDGES_PER_ELEMENT
         d = grid%edge(grid%element(elemlid)%edge(j))%degree
         if (d > 1) nrsend(to) = nrsend(to) + grid%nsoln*(d-1)
      end do
      d = grid%element(elemlid)%degree
      if (d > 2) nrsend(to) = nrsend(to) + grid%nsoln*element_dof(d)
      if (grid%oldsoln_exists) then
         gid = elemgid
         do
            elem = hash_decode_key(gid,grid%elem_hash)
            if (grid%element(elem)%oldleaf) exit
            if (grid%element(elem)%level == 1) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("failed to find element with old solution")
               stop
            endif
            gid = gid/MAX_CHILD
         end do
         nisend(to) = nisend(to) + KEY_SIZE
         nisend(to) = nisend(to) + 3
         if (associated_elem_oldsoln(grid,elem)) then
            nrsend(to) = nrsend(to) + size_elem_oldsoln(grid,elem)
         endif
         do j=1,EDGES_PER_ELEMENT
            edge = grid%element(elem)%edge(j)
            nisend(to) = nisend(to) + 3
            if (associated_edge_oldsoln(grid,edge)) then
               nrsend(to) = nrsend(to) + size_edge_oldsoln(grid,edge)
            endif
         end do
         do j=1,VERTICES_PER_ELEMENT
            vert = grid%element(elem)%vertex(j)
            nisend(to) = nisend(to) + 2
            if (associated_vert_oldsoln(grid)) then
               nrsend(to) = nrsend(to) + size_vert_oldsoln(grid,dim=COMP_DIM) * &
                                         size_vert_oldsoln(grid,dim=EIGEN_DIM)
            endif
         end do
      endif
   end do ! next export gid

! allocate messages

   allocate(isend(sum(nisend)),rsend(sum(nrsend)),stat=allocstat)
   if (allocstat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("allocation failed in redistribute",procs=procs)
      return
   endif

! start the index counters at the beginning of each segment

   iind(1) = 1
   rind(1) = 1
   do i=2,nproc
      iind(i) = iind(i-1) + nisend(i-1)
      rind(i) = rind(i-1) + nrsend(i-1)
   end do

! for each element on the export list ...

   do i=1,size(export_gid)
      elemgid = export_gid(i)
      elemlid = export_lid(i)
      to = export_part(i)
      if (.not. need_alltoall) then
         to = inv_neighproc(to)
      endif

! pack the gid

      call hash_pack_key(elemgid,isend,iind(to))
      iind(to) = iind(to) + KEY_SIZE

! pack the gids and new owners of the neighbors

      call get_all_neighbors(grid,elemlid,allneigh)
      isend(iind(to)) = size(allneigh)
      iind(to) = iind(to) + 1
      do j=1,size(allneigh)
         if (allneigh(j) == BOUNDARY) then
            isend(iind(to):iind(to)+KEY_SIZE) = 0
            isend(iind(to)) = BOUNDARY
            iind(to) = iind(to) + KEY_SIZE + 1
         else
            call hash_pack_key(grid%element(allneigh(j))%gid,isend,iind(to))
            iind(to) = iind(to) + KEY_SIZE
            if (grid%element(allneigh(j))%owner /= UNKNOWN_OWNER) then
               isend(iind(to)) = grid%element(allneigh(j))%owner
            else
               isend(iind(to)) = newowner(allneigh(j))
            endif
            iind(to) = iind(to) + 1
         endif
      end do
      deallocate(allneigh)

! pack the edge and element degrees

      isend(iind(to):iind(to)+3)=(/grid%edge(grid%element(elemlid)%edge)%degree, &
                                   grid%element(elemlid)%degree /)
      iind(to) = iind(to) + 4

! pack the error indicators

      rsend(rind(to):rind(to)+nerrind-1) = grid%element_errind(elemlid,:)
      rind(to) = rind(to) + nerrind

! pack the vertex solutions

      do j=1,VERTICES_PER_ELEMENT
         rsend(rind(to):rind(to)+grid%nsoln-1) = &
            reshape(get_vert_soln(grid,grid%element(elemlid)%vertex(j), &
                    colon,colon),(/grid%nsoln/))
         rind(to) = rind(to) + grid%nsoln
      end do

! pack the edge solutions

      do j=1,EDGES_PER_ELEMENT
         d = grid%edge(grid%element(elemlid)%edge(j))%degree
         if (d > 1) then
            do k=1,grid%system_size
               do l=1,max(1,grid%num_eval)
                  rsend(rind(to):rind(to)+d-2) = &
                    get_edge_soln(grid,grid%element(elemlid)%edge(j), &
                                  (/1,d-1/),k,l)
                  rind(to) = rind(to) + d-1
               end do
            end do
         endif
      end do

! pack the element solutions

      d = grid%element(elemlid)%degree
      if (d > 2) then
         do j=1,grid%system_size
            do k=1,max(1,grid%num_eval)
               rsend(rind(to):rind(to)+element_dof(d)-1) = &
                   get_elem_soln(grid,elemlid,(/1,element_dof(d)/),j,k)
               rind(to) = rind(to) + element_dof(d)
            end do
         end do
      endif

! if an old solution has been saved, for each export element find the element
! that has the old solution for that area (it can be the element or an ancestor)

      if (grid%oldsoln_exists) then

         gid = elemgid
         do
            elem = hash_decode_key(gid,grid%elem_hash)
            if (grid%element(elem)%oldleaf) exit
            if (grid%element(elem)%level == 1) then
               ierr = PHAML_INTERNAL_ERROR
               call fatal("failed to find element with old solution")
               stop
            endif
            gid = gid/MAX_CHILD
         end do

! pack the gid of the element that has the old solution

         call hash_pack_key(gid,isend,iind(to))
         iind(to) = iind(to) + KEY_SIZE

! pack the element old solution and its dimensions

         if (size_elem_oldsoln(grid,elem)/=0) then
            isend(iind(to)  ) = size_elem_oldsoln(grid,elem,dim=BASIS_DIM)
            isend(iind(to)+1) = size_elem_oldsoln(grid,elem,dim=COMP_DIM)
            isend(iind(to)+2) = size_elem_oldsoln(grid,elem,dim=EIGEN_DIM)
            iind(to) = iind(to) + 3
            do j=1,size_elem_oldsoln(grid,elem,dim=BASIS_DIM)
               do k=1,size_elem_oldsoln(grid,elem,dim=COMP_DIM)
                  do l=1,size_elem_oldsoln(grid,elem,dim=EIGEN_DIM)
                     rsend(rind(to)) = get_elem_oldsoln(grid,elem,j,k,l)
                     rind(to) = rind(to) + 1
                   end do
               end do
            end do
         else
            isend(iind(to)  ) = 0
            isend(iind(to)+1) = 0
            isend(iind(to)+2) = 0
            iind(to) = iind(to) + 3
         endif

! pack the edge old solution and its dimensions

         do j=1,EDGES_PER_ELEMENT
            edge = grid%element(elem)%edge(j)
            if (size_edge_oldsoln(grid,edge)/=0) then
               isend(iind(to)  ) = size_edge_oldsoln(grid,edge,dim=BASIS_DIM)
               isend(iind(to)+1) = size_edge_oldsoln(grid,edge,dim=COMP_DIM)
               isend(iind(to)+2) = size_edge_oldsoln(grid,edge,dim=EIGEN_DIM)
               iind(to) = iind(to) + 3
               do k=1,size_edge_oldsoln(grid,edge,dim=BASIS_DIM)
                  do l=1,size_edge_oldsoln(grid,edge,dim=COMP_DIM)
                     do m=1,size_edge_oldsoln(grid,edge,dim=EIGEN_DIM)
                        rsend(rind(to)) = get_edge_oldsoln(grid,edge,k,l,m)
                        rind(to) = rind(to) + 1
                     end do
                  end do
               end do
            else
               isend(iind(to)  ) = 0
               isend(iind(to)+1) = 0
               isend(iind(to)+2) = 0
               iind(to) = iind(to) + 3
            endif
         end do

! pack the vertex old solution and its dimensions

         do j=1,VERTICES_PER_ELEMENT
            vert = grid%element(elem)%vertex(j)
            if (size_vert_oldsoln(grid)/=0) then
               isend(iind(to)  ) = size_vert_oldsoln(grid,dim=COMP_DIM)
               isend(iind(to)+1) = size_vert_oldsoln(grid,dim=EIGEN_DIM)
               iind(to) = iind(to) + 2
               do k=1,isend(iind(to)-2)
                  do l=1,isend(iind(to)-1)
                     rsend(rind(to)) = get_vert_oldsoln(grid,vert,k,l)
                     rind(to) = rind(to) + 1
                  end do
               end do
            else
               isend(iind(to)  ) = 0
               isend(iind(to)+1) = 0
               iind(to) = iind(to) + 2
            endif
         end do

      endif ! old solution exists

   end do ! next export gid

! exchange information with the other processors

   call start_watch((/cpdistribute,ctdistribute/))
   if (.not. need_alltoall) then
      call phaml_neighbor_alltoall(procs,isend,nisend,irecv,nirecv,530)
      call phaml_neighbor_alltoall(procs,rsend,nrsend,rrecv,nrrecv,531)
   else
      call phaml_alltoall(procs,isend,nisend,irecv,nirecv,530)
      call phaml_alltoall(procs,rsend,nrsend,rrecv,nrrecv,531)
   endif
   call stop_watch((/cpdistribute,ctdistribute/))

   deallocate(isend,rsend,stat=allocstat)
   if (associated(neighproc)) deallocate(neighproc)

! derefine elements no longer needed

   call delete_unowned_elements(grid,refine_control,solver_control)

! unpack the elements received

   iindex = 1
   rindex = 1
   if (.not. need_alltoall) then
      limit = nneigh
   else
      limit = nproc
   endif
   do part=1,limit
      call unpack_elements(grid,procs,refine_control,solver_control,irecv, &
                           iindex,rrecv,rindex,sum(nirecv(1:part)), &
                           desired_level,desired_degree,was_hcoarsened)
   end do

! clean up oldleaf so each point in the domain has only one oldleaf

   if (grid%oldsoln_exists) then
      call cleanup_oldleaf(grid)
   endif

! For any elements in the export halo that are not in the new halo, set the
! owner to unknown.

   do i=1,nexport_halo
      elem = export_halo(i)
      if (hash_decode_key(grid%element(elem)%gid,grid%elem_hash)/=elem) then
         cycle ! elem was deleted
      endif
      if (grid%element(elem)%owner /= UNKNOWN_OWNER .and. &
          grid%element(elem)%owner /= myproc) then
         call get_all_neighbors(grid,elem,allneigh)
         do j=1,size(allneigh)
            if (allneigh(j) /= BOUNDARY) then
               if (grid%element(allneigh(j))%owner == myproc) exit
            endif
         end do
         if (j == size(allneigh)+1) then
            grid%element(elem)%owner = UNKNOWN_OWNER
         endif
         deallocate(allneigh)
      endif
   end do

! Likewise for elements in the export list.

   do i=1,size(export_lid)
      elem = export_lid(i)
      if (hash_decode_key(grid%element(elem)%gid,grid%elem_hash)/=elem) then
         cycle ! elem was deleted
      endif
      if (grid%element(elem)%owner /= UNKNOWN_OWNER .and. &
          grid%element(elem)%owner /= myproc) then
         call get_all_neighbors(grid,elem,allneigh)
         do j=1,size(allneigh)
            if (allneigh(j) /= BOUNDARY) then
               if (grid%element(allneigh(j))%owner == myproc) exit
            endif
         end do
         if (j == size(allneigh)+1) then
            grid%element(elem)%owner = UNKNOWN_OWNER
         endif
         deallocate(allneigh)
      endif
   end do

! Find the neighbor processors as owners of halo elements, and set them in procs

   call get_halo(grid,halo,nhalo,outer=.true.)

   allocate(is_neighbor(num_proc(procs)),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("redistribute: allocation failed",procs=procs)
      stop
   endif

   is_neighbor = .false.
   do i=1,nhalo
      if (grid%element(halo(i))%owner <= 0) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("redistribute: unknown owner of halo element")
         stop
      endif
      is_neighbor(grid%element(halo(i))%owner) = .true.
   end do

   allocate(neighbor_procs(count(is_neighbor)),stat=astat)
   if (astat /= 0) then
      ierr = ALLOC_FAILED
      call fatal("redistribute: allocation failed",procs=procs)
      stop
   endif

   j=0
   do i=1,num_proc(procs)
      if (is_neighbor(i)) then
         j = j+1
         neighbor_procs(j) = i
      endif
   end do

   call set_neighbor_procs(procs,neighbor_procs)

   deallocate(is_neighbor,neighbor_procs,inv_neighproc,export_halo)

! free memory

   if (associated(irecv)) deallocate(irecv,stat=allocstat)
   if (associated(rrecv)) deallocate(rrecv,stat=allocstat)
   deallocate(nirecv,nrrecv,nisend,nrsend,iind,rind,export_lid)

else ! skip_comm

! derefine elements no longer needed

   call delete_unowned_elements(grid,refine_control,solver_control)

endif ! .not. skip_comm

! stop timing the redistribution process

call stop_watch((/pdistribute,tdistribute/))

end subroutine redistribute

!          ---------------
subroutine unpack_elements(grid,procs,refine_control,solver_control,irecv, &
                           iindex,rrecv,rindex,end_of_message, &
                           desired_level,desired_degree,was_hcoarsened)
!          ---------------

!----------------------------------------------------
! This routine unpacks the elements that were exported to this processor.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(proc_info), intent(in) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control
integer, intent(in) :: irecv(:)
integer, intent(inout) :: iindex
real(my_real), intent(in) :: rrecv(:)
integer, intent(inout) :: rindex
integer, intent(in) :: end_of_message
integer, pointer, optional :: desired_level(:), desired_degree(:,:)
logical(small_logical), pointer, optional :: was_hcoarsened(:)
!----------------------------------------------------
! Local variables:

integer :: nerrind, errcode, j, k, l, m, neighlid, elemlid, d, d1, d2, d3, &
           elem, edge, vert, nneigh
type(hash_key) :: gid, elemgid, neighgid
!----------------------------------------------------
! Begin executable code

nerrind = max(1,grid%num_eval)
do while (iindex < end_of_message)
   elemgid = hash_unpack_key(irecv,iindex)
   iindex = iindex + KEY_SIZE

! create the element, and any ancestors required by it

   call create_element(grid,elemgid,refine_control,solver_control,errcode, &
                       desired_level,desired_degree,was_hcoarsened)
   if (errcode /= 0) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("grid full during redistribution",procs=procs)
      stop
   endif

! create the three neighbors, and set their owners

   nneigh = irecv(iindex)
   iindex = iindex + 1
   do j=1,nneigh
      if (irecv(iindex) == BOUNDARY) then
         iindex = iindex + KEY_SIZE + 1
      else
         neighgid = hash_unpack_key(irecv,iindex)
         iindex = iindex + KEY_SIZE
         call create_element(grid,neighgid,refine_control, &
                             solver_control,errcode,desired_level, &
                             desired_degree,was_hcoarsened)
         if (errcode /= 0) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("redistribute: grid full",procs=procs)
            stop
         endif
         neighlid = hash_decode_key(neighgid,grid%elem_hash)
         if (neighlid == HASH_NOT_FOUND) then
            ierr = PHAML_INTERNAL_ERROR
            call fatal("element just created doesn't have a neighlid in redistribute", &
                       procs=procs)
            stop
         endif
         if (irecv(iindex) /= UNKNOWN_OWNER) then
            grid%element(neighlid)%owner = irecv(iindex)
         endif
         iindex = iindex + 1
      endif
   end do

! set myself as the owner

   elemlid = hash_decode_key(elemgid,grid%elem_hash)
   if (elemlid == HASH_NOT_FOUND) then
      ierr = PHAML_INTERNAL_ERROR
      call fatal("element just created doesn't have an elemlid in redistribute", &
                 procs=procs)
      stop
   endif
   grid%element(elemlid)%owner = my_proc(procs)
   if (.not. grid%element(elemlid)%iown) then
      grid%element(elemlid)%iown = .true.
      call own_parent(grid,elemgid,procs)
      grid%nelem_leaf_own = grid%nelem_leaf_own + 1
      if (grid%element(elemlid)%degree >= 3) then
         grid%dof_own = grid%dof_own + grid%system_size * &
            element_dof(grid%element(elemlid)%degree)
      endif
      do j=1,VERTICES_PER_ELEMENT
         if (grid%vertex(grid%element(elemlid)%vertex(j))%assoc_elem == elemlid) then
            grid%nvert_own = grid%nvert_own + 1
            grid%dof_own = grid%dof_own + grid%system_size
         endif
      end do
      do j=1,EDGES_PER_ELEMENT
         if (grid%edge(grid%element(elemlid)%edge(j))%assoc_elem == elemlid) then
            grid%nedge_leaf_own = grid%nedge_leaf_own + 1
            if (grid%edge(grid%element(elemlid)%edge(j))%degree >= 2) then
               grid%dof_own = grid%dof_own + grid%system_size * &
                     (grid%edge(grid%element(elemlid)%edge(j))%degree - 1)
            endif
         endif
      end do
   endif

! set the edge and element degrees

   if (grid%element(elemlid)%isleaf) then
      do while (grid%element(elemlid)%degree < irecv(iindex+3))
         call p_refine_elem(grid,elemlid,refine_control)
      end do
   endif

! error indicators

   grid%element_errind(elemlid,:) = rrecv(rindex:rindex+nerrind-1)
   rindex = rindex + nerrind

! set the solution values of this element

! NOTE: for edges, it is possible that the current degree does not agree with
! the degree that was sent, because the p refinement of a neighboring element
! to bring up the degree may not have occured yet.  So use the minimum of
! edge degree and what was sent as the number to copy (the others will be
! copied when the neighbor is done), and increment the message pointer by
! the number that was sent.  This does not occur with vertices and elements.

! vertex solutions

   do j=1,VERTICES_PER_ELEMENT
      call set_vert_soln(grid,grid%element(elemlid)%vertex(j),colon, &
             colon,reshape(rrecv(rindex:rindex+grid%nsoln-1), &
             (/grid%system_size,max(1,grid%num_eval)/)))
      rindex = rindex + grid%nsoln
   end do

! edge solutions

   do j=1,EDGES_PER_ELEMENT
      d = min(grid%edge(grid%element(elemlid)%edge(j))%degree, &
              irecv(iindex+j-1))
      do k=1,grid%system_size
         do l=1,max(1,grid%num_eval)
            if (d > 1) then
               call set_edge_soln(grid,grid%element(elemlid)%edge(j), &
                                  (/1,d-1/),k,l,rrecv(rindex:rindex+d-2))
            endif
            rindex = rindex + irecv(iindex+j-1)-1
         end do
      end do
   end do

! element solutions

   d = min(grid%element(elemlid)%degree,irecv(iindex+3))
   if (d > 2) then
      do j=1,grid%system_size
         do k=1,max(1,grid%num_eval)
            call set_elem_soln(grid,elemlid,(/1,element_dof(d)/),j,k, &
                               rrecv(rindex:rindex+element_dof(d)-1))
            rindex = rindex + element_dof(irecv(iindex+3))
         end do
      end do
   endif

   iindex = iindex + 4

! if an old solution exists, set the old solution values sent

   if (grid%oldsoln_exists) then

      gid = hash_unpack_key(irecv,iindex)
      iindex = iindex + KEY_SIZE
      elem = hash_decode_key(gid,grid%elem_hash)
      if (elem == HASH_NOT_FOUND) then
         ierr = PHAML_INTERNAL_ERROR
         call fatal("element for old solution not found in redistribute",&
                    procs=procs)
         stop
      endif

! element old solution

      if (irecv(iindex) /= 0) then
         d1 = irecv(iindex  )
         d2 = irecv(iindex+1)
         d3 = irecv(iindex+2)
         if (associated_elem_oldsoln(grid,elem)) then
            if (size_elem_oldsoln(grid,elem,dim=BASIS_DIM) /= d1 .or. &
                size_elem_oldsoln(grid,elem,dim=COMP_DIM) /= d2 .or. &
                size_elem_oldsoln(grid,elem,dim=EIGEN_DIM) /= d3) then
               call deallocate_elem_oldsoln(grid,elem)
            endif
         endif
         if (.not. associated_elem_oldsoln(grid,elem)) then
            call allocate_elem_oldsoln(grid,elem,size_basis=d1, &
                                       size_comp=d2,size_eigen=d3)
         endif
         do j=1,d1
            do k=1,d2
               do l=1,d3
                  call set_elem_oldsoln(grid,elem,j,k,l,rrecv(rindex))
                  rindex = rindex + 1
               end do
            end do
         end do
      else
         if (associated_elem_oldsoln(grid,elem)) then
            call deallocate_elem_oldsoln(grid,elem)
         endif
      endif
      iindex = iindex + 3

! edge old solution

      do j=1,EDGES_PER_ELEMENT
         edge = grid%element(elem)%edge(j)
         if (irecv(iindex) /= 0) then
            d1 = irecv(iindex  )
            d2 = irecv(iindex+1)
            d3 = irecv(iindex+2)
            if (associated_edge_oldsoln(grid,edge)) then
               if (size_edge_oldsoln(grid,edge,dim=BASIS_DIM) /= d1 .or. &
                   size_edge_oldsoln(grid,edge,dim=COMP_DIM) /= d2 .or. &
                   size_edge_oldsoln(grid,edge,dim=EIGEN_DIM) /= d3) then
                  call deallocate_edge_oldsoln(grid,edge)
               endif
            endif
            if (.not. associated_edge_oldsoln(grid,edge)) then
               call allocate_edge_oldsoln(grid,edge,size_basis=d1, &
                                          size_comp=d2,size_eigen=d3)
            endif
            do k=1,d1
               do l=1,d2
                  do m=1,d3
                     call set_edge_oldsoln(grid,edge,k,l,m,rrecv(rindex))
                     rindex = rindex + 1
                  end do
               end do
            end do
         else
            if (associated_edge_oldsoln(grid,edge)) then
               call deallocate_edge_oldsoln(grid,edge)
            endif
         endif
         iindex = iindex + 3
      end do

! vertex old solution

      do j=1,VERTICES_PER_ELEMENT
         vert = grid%element(elem)%vertex(j)
         if (irecv(iindex) /= 0) then
            d1 = irecv(iindex  )
            d2 = irecv(iindex+1)
            do k=1,d1
               do l=1,d2
                  call set_vert_oldsoln(grid,vert,k,l,rrecv(rindex))
                  rindex = rindex + 1
               end do
            end do
         endif
         iindex = iindex + 2
      end do

! set this element as an oldleaf

      grid%element(elem)%oldleaf = .true.

   endif ! old solution exists
end do ! next imported element

end subroutine unpack_elements

!          ------------------
subroutine update_halo_owners(grid,procs,export_lid,export_part,inv_neighproc, &
                              nneigh,need_alltoall,export_halo,nexport_halo)
!          ------------------

!----------------------------------------------------
! This routine informs neighboring processors of the new owner of elements that
! this processor is exporting and are in the neighbor's halo.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(proc_info), intent(inout) :: procs
integer, intent(in) :: export_lid(:), export_part(:), inv_neighproc(:), nneigh
logical, intent(in) :: need_alltoall
integer, intent(out) :: export_halo(:), nexport_halo
!----------------------------------------------------
! Local variables:

integer :: i, j, myproc, limit, elem, to, astat
integer :: tempowner(grid%biggest_elem)
integer, pointer :: allneigh(:), irecv(:)
integer, allocatable :: nsend(:), nrecv(:), iind(:), isend(:)

!----------------------------------------------------
! Begin executable code

myproc = my_proc(procs)

! Temporarily set the owner of exported elements to the new owner

do i=1,size(export_lid)
   grid%element(export_lid(i))%owner = export_part(i)
end do

! Determine the size of the message segments and allocate the message.
! Also make a list of elements in the halo that share a vertex with an export
! element.

if (need_alltoall) then
   limit = num_proc(procs)
else
   limit = nneigh
endif
allocate(nsend(limit),nrecv(limit),iind(limit),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("update_halo_owners: allocation failed",procs=procs)
   stop
endif

nsend = 0
nexport_halo = 0
do i=1,size(export_lid)
   elem = export_lid(i)
   call get_all_neighbors(grid,elem,allneigh)
   do j=1,size(allneigh)
      if (allneigh(j) /= BOUNDARY) then
         if (grid%element(allneigh(j))%owner /= UNKNOWN_OWNER .and. &
             grid%element(allneigh(j))%owner /= myproc) then
            to = grid%element(allneigh(j))%owner
            if (.not. need_alltoall) to = inv_neighproc(to)
            nsend(to) = nsend(to) + 1
            nexport_halo = nexport_halo + 1
            export_halo(nexport_halo) = allneigh(j)
         endif
      endif
   end do
   deallocate(allneigh)
end do
nsend = nsend*(KEY_SIZE+1)

allocate(isend(sum(nsend)),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("update_halo_owners: allocation failed",procs=procs)
   stop
endif

! Pass through the export list checking each neighbor of its entries to see if
! the owner is a different processor.  If so, add the element and new owner
! to the message for the neighbor processor.
! Also set the new owner for those elements that would have been in a message
! to myself in a temporary array to be copied later.

iind(1) = 1
do i=2,limit
   iind(i) = iind(i-1) + nsend(i-1)
end do

do i=1,size(export_lid)
   elem = export_lid(i)
   tempowner(elem) = UNKNOWN_OWNER
   call get_all_neighbors(grid,elem,allneigh)
   do j=1,size(allneigh)
      if (allneigh(j) /= BOUNDARY) then
         if (grid%element(allneigh(j))%owner /= UNKNOWN_OWNER) then
            if (grid%element(allneigh(j))%owner == myproc) then
               tempowner(elem) = export_part(i)
            else
               to = grid%element(allneigh(j))%owner
               if (.not. need_alltoall) to = inv_neighproc(to)
               call hash_pack_key(grid%element(elem)%gid,isend,iind(to))
               isend(iind(to)+KEY_SIZE) = export_part(i)
               iind(to) = iind(to) + KEY_SIZE+1
            endif
         endif
      endif
   end do
   deallocate(allneigh)
end do

! Copy the temporary array that contains the "message to myself" to set the
! owner of export elements that are neighbors of elements that I have, or to
! return the owner to UNKNOWN_OWNER.

do i=1,size(export_lid)
   grid%element(export_lid(i))%owner = tempowner(export_lid(i))
end do

! Exchange messages

call start_watch((/cpdistribute,ctdistribute/))
if (need_alltoall) then
   call phaml_alltoall(procs,isend,nsend,irecv,nrecv,591)
else
   call phaml_neighbor_alltoall(procs,isend,nsend,irecv,nrecv,591)
endif
call stop_watch((/cpdistribute,ctdistribute/))

! Change the owner of received elements that I have

do i=1,sum(nrecv),KEY_SIZE+1
   elem = hash_decode_key(hash_unpack_key(irecv,i),grid%elem_hash)
   if (elem /= HASH_NOT_FOUND) then
      grid%element(elem)%owner = irecv(i+KEY_SIZE)
   endif
end do

deallocate(isend,nsend,irecv,nrecv,iind)

end subroutine update_halo_owners

!          -----------------------
subroutine delete_unowned_elements(grid,refcont,solver_control)
!          -----------------------

!----------------------------------------------------
! This routine removes any elements that are not owned by this partition
! and are not needed for compatibility.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(refine_options), intent(in) :: refcont
type(solver_options), intent(in) :: solver_control
!----------------------------------------------------
! Local variables:

integer :: elem
logical :: owned_child
!----------------------------------------------------
! Begin executable code
!

elem = grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   call delete_unowned_elements_recur(grid,elem,owned_child,refcont, &
                                      solver_control)
   elem = grid%element(elem)%next
end do

end subroutine delete_unowned_elements

!                    -----------------------------
recursive subroutine delete_unowned_elements_recur(grid,elem,owned_child, &
                                                   refcont,solver_control)
!                    -----------------------------

!----------------------------------------------------
! This routine does the work of delete_unowned_elements
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: elem
logical, intent(out) :: owned_child
type(refine_options), intent(in) :: refcont
type(solver_options), intent(in) :: solver_control

!----------------------------------------------------
! Local variables:

integer :: child(MAX_CHILD), i, mate, errcode, allc(MAX_CHILD)
logical :: this_owned_child
type(hash_key) :: mate_gid
!----------------------------------------------------
! Begin executable code

allc = ALL_CHILDREN
child = get_child_lid(grid%element(elem)%gid,allc,grid%elem_hash)

! if this is a leaf, return with an indication of whether it is owned by
! this partition

if (child(1) == NO_CHILD) then
   owned_child = .not. grid%element(elem)%owner == UNKNOWN_OWNER
else

! otherwise, see if any of the descendents are owned

   owned_child = .false.
   do i=1,MAX_CHILD
      if (child(i) == NO_CHILD) cycle
      call delete_unowned_elements_recur(grid,child(i),this_owned_child, &
                                         refcont,solver_control)
      owned_child = owned_child .or. this_owned_child
   end do

! also check the descendents of the mate

   if (.not. owned_child) then
      mate_gid = grid%element(elem)%mate
      if (.not. (mate_gid == BOUNDARY)) then
         mate = hash_decode_key(mate_gid,grid%elem_hash)
         if (mate /= HASH_NOT_FOUND) then
            child = get_child_lid(grid%element(mate)%gid,allc,grid%elem_hash)
            do i=1,MAX_CHILD
              if (child(i) == NO_CHILD) cycle
              call delete_unowned_elements_recur(grid,child(i), &
                                                 this_owned_child,refcont, &
                                                 solver_control)
              owned_child = owned_child .or. this_owned_child
            end do
         endif
      endif
   endif

! if not, derefine this element

   if (.not. owned_child) then
      call h_coarsen_element(grid,elem,errcode,refcont,solver_control)
   endif

endif

end subroutine delete_unowned_elements_recur

!                    -------------
recursive subroutine disown_parent(grid,child)
!                    -------------

!----------------------------------------------------
! This routine removes ownership of all ancestors of child
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(hash_key), intent(in) :: child
!----------------------------------------------------
! Local variables:

type(hash_key) :: parent
integer :: lid
!----------------------------------------------------
! Begin executable code

parent = child/MAX_CHILD
lid = hash_decode_key(parent,grid%elem_hash)
if (lid == HASH_NOT_FOUND) return

if (grid%element(lid)%iown) then
   grid%element(lid)%iown = .false.
   grid%element(lid)%owner = UNKNOWN_OWNER
   call disown_parent(grid,parent)
endif

end subroutine disown_parent

!                    ----------
recursive subroutine own_parent(grid,child,procs)
!                    ----------

!----------------------------------------------------
! This routine adds ownership of ancestors of child for which
! all descendents are owned
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(hash_key), intent(in) :: child
type(proc_info), intent(in) :: procs
!----------------------------------------------------
! Local variables:

type(hash_key) :: parent
integer :: children(MAX_CHILD), i, lid, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

parent = child/MAX_CHILD
lid = hash_decode_key(parent,grid%elem_hash)
if (lid == HASH_NOT_FOUND) return
allc = ALL_CHILDREN
children = get_child_lid(parent,allc,grid%elem_hash)

grid%element(lid)%iown = .true.
do i=1,MAX_CHILD
   if (children(i) /= NO_CHILD) then
      grid%element(lid)%iown = grid%element(lid)%iown .and. &
                               grid%element(children(i))%iown
   endif
end do

if (grid%element(lid)%iown) grid%element(lid)%owner = my_proc(procs)

if (grid%element(lid)%iown) call own_parent(grid,parent,procs)

end subroutine own_parent

!          ---------------
subroutine cleanup_oldleaf(grid)
!          ---------------

!----------------------------------------------------
! This routine cleans up oldleaf to make sure any point has only one
! element marked as a leaf for the old solution.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
!----------------------------------------------------
! Local variables:

integer :: elem
logical :: dum1, dum2
!----------------------------------------------------
! Begin executable code

! pass through all elements of the initial grid calling the worker recursive
! routine

elem = grid%head_level_elem(1)
do while (elem /= END_OF_LIST)
   call cleanup_oldleaf_recur(grid,elem,dum1,dum2)
   elem = grid%element(elem)%next
end do

end subroutine cleanup_oldleaf

!                    ---------------------
recursive subroutine cleanup_oldleaf_recur(grid,parent,any_oldleaf,all_oldleaf)
!                    ---------------------

!----------------------------------------------------
! This routine does the work of cleanup_oldleaf
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: parent
logical, intent(out) :: any_oldleaf, all_oldleaf
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
logical :: any_child, all_child
!----------------------------------------------------
! Begin executable code

! any_oldleaf for parent is true if any of its area is an oldleaf.
! all_oldleaf for parent is true if all of its area is covered by oldleafs

allc = ALL_CHILDREN
children = get_child_lid(grid%element(parent)%gid,allc,grid%elem_hash)

! if this is a leaf, then any_oldleaf and all_oldleaf are the same as oldleaf

if (children(1) == NO_CHILD) then
   any_oldleaf = grid%element(parent)%oldleaf
   all_oldleaf = grid%element(parent)%oldleaf

! otherwise, any_oldleaf is the or of the children and all_oldleaf is the and

else

   any_oldleaf = .false.
   all_oldleaf = .true.
   do i=1,MAX_CHILD
      call cleanup_oldleaf_recur(grid,children(i),any_child,all_child)
      any_oldleaf = any_oldleaf .or. any_child
      all_oldleaf = all_oldleaf .and. all_child
   end do

! if this non-leaf element is marked as an oldleaf, then if it is all_oldleaf
! unmark it as oldleaf, and if it is any_oldleaf but not all_oldleaf, then
! unmark any descendants that are oldleaf

   if (grid%element(parent)%oldleaf) then
      if (all_oldleaf) then
         grid%element(parent)%oldleaf = .false.
      elseif (any_oldleaf) then
         do i=1,MAX_CHILD
            call unset_oldleaf(grid,children(i))
         end do
      endif
      any_oldleaf = .true.
      all_oldleaf = .true.
   endif

endif

end subroutine cleanup_oldleaf_recur

!                    -------------
recursive subroutine unset_oldleaf(grid,parent)
!                    -------------

!----------------------------------------------------
! This routine unsets oldleaf in all descendents of parent and parent
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
integer, intent(in) :: parent
!----------------------------------------------------
! Local variables:

integer :: children(MAX_CHILD), i, allc(MAX_CHILD)
!----------------------------------------------------
! Begin executable code

grid%element(parent)%oldleaf = .false.

allc = ALL_CHILDREN
children = get_child_lid(grid%element(parent)%gid,allc,grid%elem_hash)

if (children(1) /= NO_CHILD) then
   do i=1,MAX_CHILD
      call unset_oldleaf(grid,children(i))
   end do
endif

end subroutine unset_oldleaf

!          ---------------
subroutine set_halo_owners(grid,procs,refine_control,solver_control)
!          ---------------

!----------------------------------------------------
! This routine sets the owners of the halo elements, and elements that this
! processor owns.
!----------------------------------------------------

!----------------------------------------------------
! Dummy arguments

type(grid_type), intent(inout) :: grid
type(proc_info), intent(inout) :: procs
type(refine_options), intent(in) :: refine_control
type(solver_options), intent(in) :: solver_control

!----------------------------------------------------
! Local variables:

integer :: myproc, nproc, lev, elem, nhalo, astat, i, j, k, proc, &
           errcode, halo(grid%nelem_leaf)
logical :: inhalo
integer, pointer :: irecv(:), allneigh(:)
integer, allocatable :: isend(:), nrecv(:), neighbor_procs(:)
logical, allocatable :: is_neighbor(:)
type(hash_key) :: gid

!----------------------------------------------------
! Begin executable code

! consider this time to be part of the redistribution process

call start_watch((/pdistribute,tdistribute/))

myproc = my_proc(procs)
nproc  = num_proc(procs)

! Set owner of unowned elements to UNKNOWN_OWNER and owned elements to my_proc

do lev=1,grid%nlev
   elem = grid%head_level_elem(lev)
   do while (elem /= END_OF_LIST)
      if (grid%element(elem)%iown) then
         grid%element(elem)%owner = myproc
      else
         grid%element(elem)%owner = UNKNOWN_OWNER
      endif
      elem = grid%element(elem)%next
   end do
end do

! Get the inner halo, i.e. a list of leaf elements that I own and have a
! vertex on the partition boundary

call get_halo(grid,halo,nhalo,inner=.true.)

allocate(isend(nhalo*KEY_SIZE),nrecv(nproc),stat=astat)
if (astat /= 0) then
   ierr = ALLOC_FAILED
   call fatal("set_halo_owners: allocation failed",procs=procs)
   stop
endif

! Copy the halo to one with gids

i = 0
do j=1,nhalo
   call hash_pack_key(grid%element(halo(j))%gid,isend,i+1)
   i = i + KEY_SIZE
end do

! Exchange lists with all processors

call start_watch((/cpdistribute,ctdistribute/))
call phaml_alltoall(procs,isend,i,irecv,nrecv,590)
call stop_watch((/cpdistribute,ctdistribute/))

! For each element received (except my own), set the owner to the sender if it
! is in my outer halo, i.e. if it has a neighbor that I own.  If it doesn't
! exist, create it.

i = 0
do proc=1,nproc
   if (proc == myproc) then
      i = i + nrecv(proc)
   else
      do j=1,nrecv(proc),KEY_SIZE
         gid = hash_unpack_key(irecv,i+1)
         elem = hash_decode_key(gid,grid%elem_hash)
         if (elem == HASH_NOT_FOUND) then
            inhalo = .true.
            call create_element(grid,gid,refine_control,solver_control, &
                                errcode,inhalo=inhalo)
            elem = hash_decode_key(gid,grid%elem_hash)
         endif
         if (elem /= HASH_NOT_FOUND) then
            call get_all_neighbors(grid,elem,allneigh)
            do k=1,size(allneigh)
               if (allneigh(k) /= BOUNDARY) then
                  if (grid%element(allneigh(k))%owner==myproc) then
                     grid%element(elem)%owner = proc
                     exit
                  endif
               endif
            end do
            if (associated(allneigh)) deallocate(allneigh)
         endif
         i = i + KEY_SIZE
      end do
   endif
end do

deallocate(isend,irecv,nrecv)

! Delete unneeded elements that are not owned

call delete_unowned_elements(grid,refine_control,solver_control)

! Find the neighbor processors as owners of halo elements, and set them in procs

call get_halo(grid,halo,nhalo,outer=.true.)

allocate(is_neighbor(num_proc(procs)),stat=astat)

is_neighbor = .false.
do i=1,nhalo
   is_neighbor(grid%element(halo(i))%owner) = .true.
end do

allocate(neighbor_procs(count(is_neighbor)),stat=astat)

j=0
do i=1,num_proc(procs)
   if (is_neighbor(i)) then
      j = j+1
      neighbor_procs(j) = i
   endif
end do

call set_neighbor_procs(procs,neighbor_procs)

deallocate(is_neighbor,neighbor_procs)

grid%errind_up2date = .false.

! stop timing

call stop_watch((/pdistribute,tdistribute/))

end subroutine set_halo_owners

end module load_balance
